/*-
 * Copyright (c) 2006 John Birrell jb@freebsd.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <errno.h>
#include <fcntl.h>
#include <libelf.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "_libelf.h"

static const char *
ei_class_str(u_int ei_class)
{
	switch (ei_class) {
	case ELFCLASSNONE:	return("ELFCLASSNONE");
	case ELFCLASS32:	return("ELFCLASS32");
	case ELFCLASS64:	return("ELFCLASS64");
	default:		return (NULL);
	}
}

static const char *
ei_data_str(u_int ei_data)
{
	switch (ei_data) {
	case ELFDATANONE:	return("ELFDATANONE");
	case ELFDATA2LSB:	return("ELFDATA2LSB");
	case ELFDATA2MSB:	return("ELFDATA2MSB");
	default:		return (NULL);
	}
}

static const char *
ei_osabi_str(u_int ei_osabi)
{
	switch (ei_osabi) {
	case ELFOSABI_NONE:		return("ELFOSABI_NONE");
	case ELFOSABI_86OPEN:		return("ELFOSABI_86OPEN");
	case ELFOSABI_AIX:		return("ELFOSABI_AIX");
	case ELFOSABI_ARM:		return("ELFOSABI_ARM");
	case ELFOSABI_FREEBSD:		return("ELFOSABI_FREEBSD");
	case ELFOSABI_HPUX:		return("ELFOSABI_HPUX");
	case ELFOSABI_HURD:		return("ELFOSABI_HURD");
	case ELFOSABI_IRIX:		return("ELFOSABI_IRIX");
	case ELFOSABI_LINUX:		return("ELFOSABI_LINUX");
	case ELFOSABI_MODESTO:		return("ELFOSABI_MODESTO");
	case ELFOSABI_NETBSD:		return("ELFOSABI_NETBSD");
	case ELFOSABI_NSK:		return("ELFOSABI_NSK");
	case ELFOSABI_OPENBSD:		return("ELFOSABI_OPENBSD");
	case ELFOSABI_OPENVMS:		return("ELFOSABI_OPENVMS");
	case ELFOSABI_SOLARIS:		return("ELFOSABI_SOLARIS");
	case ELFOSABI_STANDALONE:	return("ELFOSABI_STANDALONE");
	case ELFOSABI_TRU64:		return("ELFOSABI_TRU64");
	default:			return (NULL);
	}
}

static const char *
et_str(u_int e_type)
{
	switch (e_type) {
	case ET_NONE:	return("ET_NONE");
	case ET_REL:	return("ET_REL");
	case ET_EXEC:	return("ET_EXEC");
	case ET_DYN:	return("ET_DYN");
	case ET_CORE:	return("ET_CORE");
	default:	return (NULL);
	}
}

static const char *
em_str(u_int e_machine)
{
	switch (e_machine) {
	case EM_NONE:		return ("EM_NONE");
	case EM_386:		return ("EM_386");
	case EM_68HC12:		return ("EM_68HC12");
	case EM_68K:		return ("EM_68K");
	case EM_860:		return ("EM_860");
	case EM_88K:		return ("EM_88K");
	case EM_960:		return ("EM_960");
	case EM_ARC:		return ("EM_ARC");
	case EM_ARM:		return ("EM_ARM");
	case EM_COLDFIRE:	return ("EM_COLDFIRE");
	case EM_FR20:		return ("EM_FR20");
	case EM_H8S:		return ("EM_H8S");
	case EM_H8_300:		return ("EM_H8_300");
	case EM_H8_300H:	return ("EM_H8_300H");
	case EM_H8_500:		return ("EM_H8_500");
	case EM_IA_64:		return ("EM_IA_64");
	case EM_M32:		return ("EM_M32");
	case EM_ME16:		return ("EM_ME16");
	case EM_MIPS:		return ("EM_MIPS");
	case EM_MIPS_RS3_LE:	return ("EM_MIPS_RS3_LE");
	case EM_MIPS_X:		return ("EM_MIPS_X");
	case EM_MMA:		return ("EM_MMA");
	case EM_NCPU:		return ("EM_NCPU");
	case EM_NDR1:		return ("EM_NDR1");
	case EM_PARISC:		return ("EM_PARISC");
	case EM_PCP:		return ("EM_PCP");
	case EM_PPC:		return ("EM_PPC");
	case EM_PPC64:		return ("EM_PPC64");
	case EM_RCE:		return ("EM_RCE");
	case EM_RH32:		return ("EM_RH32");
	case EM_S370:		return ("EM_S370");
	case EM_S390:		return ("EM_S390");
	case EM_SH:		return ("EM_SH");
	case EM_SPARC:		return ("EM_SPARC");
	case EM_SPARC32PLUS:	return ("EM_SPARC32PLUS");
	case EM_SPARCV9:	return ("EM_SPARCV9");
	case EM_ST100:		return ("EM_ST100");
	case EM_STARCORE:	return ("EM_STARCORE");
	case EM_TINYJ:		return ("EM_TINYJ");
	case EM_TRICORE:	return ("EM_TRICORE");
	case EM_V800:		return ("EM_V800");
	case EM_VPP500:		return ("EM_VPP500");
	case EM_X86_64:		return ("EM_X86_64");
	default:		return (NULL);
	}
}

static const char *
ev_str(u_int e_version)
{
	switch (e_version) {
	case EV_NONE:		return ("EV_NONE");
	case EV_CURRENT:	return ("EV_CURRENT");
	default:		return (NULL);
	}
}

#define LIBELF_DUMP_NN	32
#include "libelf_dump.c"

#undef LIBELF_DUMP_NN
#define LIBELF_DUMP_NN	64
#include "libelf_dump.c"

int
elf_dump(const char *elf_file, const char *name, const char *outfile, unsigned int flags)
{
	Elf *e;
	int err = 0;
	int fd;

	if (elf_file == NULL)
		return (EINVAL);

	elf_version(EV_CURRENT);

	if ((fd = open(elf_file, O_RDONLY)) < 0) {
		fprintf(stderr, "elfdump: open: '%s' - %s\n", elf_file, strerror(errno));
		return (errno);
	}

	if ((e = elf_begin(fd, ELF_C_READ, NULL)) != NULL) {
		if (_libelf_get_elf_class(e) == ELFCLASS32)
			err = _libelf_dump32(e, name, outfile, flags);
		else
			err = _libelf_dump64(e, name, outfile, flags);

		elf_end(e);
	} else {
		err = elf_errno();
		fprintf(stderr, "Error %d in elf_begin - %s\n", err, elf_errmsg(err));
	}

	close(fd);

	return (err);
}

