/*-
 * Copyright (c) 2007 John Birrell (jb@freebsd.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include "libdwarf.h"

#define	DWD_ABBREV	(1 << 0)
#define	DWD_INFO	(1 << 1)
#define	DWD_SHSTR	(1 << 2)
#define	DWD_STR		(1 << 3)
#define	DWD_EXTRA	(1 << 4)
#define	DWD_ALL		DWD_ABBREV | DWD_INFO | DWD_SHSTR | DWD_STR

static void
usage(void)
{
	printf("Usage: dwarfdump [-abcdefilmoprsvy] [-t{afv}] [-uobjectfile] filename [...]\n");
	exit(1);
}

int
main (int argc, char *argv[])
{
	Dwarf_Debug dbg = NULL;
	Dwarf_Error error;
	int ch;
	int fd;
	int i;
	uint32_t flags = 0;

	bzero(&error, sizeof(error));

	while ((ch = getopt(argc, argv, "abcdefhilmoprsvyzt:u:")) != -1)
		switch (ch) {
		case 'a':
			flags |= DWD_ALL;
			break;
		case 'b':
			flags |= DWD_ABBREV;
			break;
		case 'c':
			break;
		case 'd':
			break;
		case 'e':
			break;
		case 'f':
			break;
		case 'h':
			flags |= DWD_SHSTR;
			break;
		case 'i':
			flags |= DWD_INFO;
			break;
		case 'l':
			break;
		case 'm':
			break;
		case 'o':
			break;
		case 'p':
			break;
		case 'r':
			break;
		case 's':
			flags |= DWD_STR;
			break;
		case 't':
			break;
		case 'u':
			break;
		case 'v':
			break;
		case 'y':
			break;
		case 'z':
			flags |= DWD_EXTRA;
			break;
		case '?':
		default:
			usage();
		}
	argc -= optind;
	argv += optind;
	if (argc == 0)
		usage();

	for (i = 0; i < argc; i++) {
		if ((fd = open(argv[i], O_RDONLY, 0)) < 0)
			err(1, "%s", argv[i]);

		if (dwarf_init(fd, DW_DLC_READ, &dbg, &error) != DWARF_E_NONE)
			errx(1, "%s", dwarf_errmsg(&error));

		if ((flags & DWD_ABBREV) != 0)
			dwarf_dump_abbrev(dbg);

		if ((flags & DWD_INFO) != 0)
			dwarf_dump_info(dbg);

		if ((flags & DWD_SHSTR) != 0)
			dwarf_dump_shstrtab(dbg);

		if ((flags & DWD_STR) != 0)
			dwarf_dump_strtab(dbg);

		if ((flags & DWD_EXTRA) != 0) {
			dwarf_dump_tree(dbg);
		}

		if (dwarf_finish(&dbg, &error) != DWARF_E_NONE)
			errx(1, "%s", dwarf_errmsg(&error));

		close(fd);
	}
	
	return 0;
}
#ifdef DOODAD
	  -a   Dumps all sections.  Same as -bcfilmoprsy -tfv.

	  -b   Dumps the .debug_abbrev section.

	  -c   Dumps the .debug_loc section.

	  -d   Dense  mode.   Each   die   information	 of   the
	       .debug_info section is printed in one-line format.
	       This option does	not imply -i.

	  -e   Ellipsis	 mode.	 Short	names  for  DW_TAG_*  and
	       DW_ATTR_*   are	 used	in  the	 output	 for  the
	       .debug_info section.

	  -f   Dumps the .debug_frame section.

	  -i   Dumps the .debug_info section.

	  -l   Dumps the .debug_line information.

	  -m   Dumps the .debug_macinfo	section.

	  -o   Dumps the .reloc_debug_*	sections.

	  -p   Dumps the .debug_pubnames section.

	  -r   Dumps the .debug_aranges	section.

	  -s   Dumps .debug_string section.

	  -ta  Same as -tfv.

	  -tf  Dumps the .debug_static_funcs section.

	  -tv  Dumps the .debug_static_vars section.

	  -uofile
	       Restricts the dumping of	 sections  to  the  named
	       compilation unit.

	  -v   Verbose mode.  Shows  more  detailed  information.
	       More  detailed  information about the .debug_frame
	       section prints if 2 or 3	-v options are given.

	  -w   Dumps the .debug_weaknames section.

	  -y   Dumps the .debug_types section.
#endif
